#!/usr/bin/ruby
require "open3"

if (ARGV[0].nil? or not File.exists?(ARGV[0])) then
    puts "A valid data file should be included as argument."
    puts " * Required data nomenclature: [prefixdirname]_[Band]/[date]_[time]_[source].txt"
    puts "   For example:                transito_S/20110404_084111_3C405.txt"
    exit(-1)
end

for arg in ARGV

    # Get filename with extention and without it
    vpath = arg.split("/")
    filename_and_ext = vpath[vpath.length-1]
    filename = filename_and_ext.split(".")
    filename = filename[0..filename.length-2].join("")

    # Get directory where the file resides and the band (which forms part of the last directory)
    directory = "./"
    band = "?"
    if (vpath.length > 1) then
        directory = vpath[0..vpath.length-2].join("/") + "/"
        last_dir = vpath[vpath.length-2]
        if (last_dir.split("_").length > 1) then
            name_parts = last_dir.split("_")
            band = name_parts[name_parts.length-1]
        end
    end


    infile = File.new(directory + filename_and_ext, "r")
    outfile = File.new(directory + filename + ".dat", "w")
    logfile = File.new(directory + filename + ".log", "w")


    date = filename_and_ext.split("_")[0]
    date = date[0..3] + "/" + date[4..5] + "/" + date[6..7]
    hms = filename_and_ext.split("_")[1]
    hms = hms[0..1] + ":" + hms[2..3] + ":" + hms[4..5]

    puts "FILE=" + directory + filename_and_ext
    logfile.write "FILE=" + directory + filename_and_ext + "\n"
    puts "DATE=" + date + "\t" + "TIME=" + hms
    logfile.write "DATE=" + date + "\t" + "TIME=" + hms + "\n"
    puts "BAND=" + band
    logfile.write "BAND=" + band + "\n"
    linenumber = 1
    header = Hash.new
    ydata = []
    infile.each { |i|
            case linenumber
                when 1
                    vline = i.split("\t")
                    header["catalog"] = vline[0].split("=")[1].chop
                    header["source"] = vline[1].split("=")[1].chop
                    puts i
                    logfile.write i
                when 2
                    vline = i.split("\t")
                    header["dec"] = vline[0].split("=")[1].to_f
                    header["ha"] = vline[1].split("=")[1].to_f
                    header["az"] = vline[2].split("=")[1].to_f
                    header["el"] = vline[3].split("=")[1].to_f
                    puts i
                    logfile.write i
                when 3
                    vline = i.split("\t")
                    header["offset_xdec"] = vline[0].split("=")[1].to_i / 1000
                    header["offset_dec"] = vline[1].split("=")[1].to_i  / 1000
                    header["offset_transit"] = vline[2].split("=")[1].to_f / 1000
                    header["offset_source_value"] = vline[3].split("=")[1].to_f
                    puts i
                    logfile.write i
                when 4
                    # Ignore (empty line)
                else
                    i = i.chop # Remove \n
                    ydata.push(i)
            end
            linenumber += 1
    }

    hadelt = (-2.0*header["offset_transit"])/(ydata.length-1)
    n = 0
    for y in ydata
        x = header["offset_transit"] + n * hadelt
        outfile.write x.to_s + "\t" + y.to_s + "\n"
        n += 1
    end

    outfile.close()
    infile.close()

# DISCARDED: The same plot is drawn with octave
#    Open3.popen3('/usr/bin/gnuplot') { |stdin, stdout, stderr| 
#        commands = "
#            set grid
#            set key top left
#            set key box
#            set nokey
#            set rmargin 10
#            set lmargin 10
#            set terminal png size 800, 600
#            set title '" + directory + filename_and_ext + "'
#            set xlabel 'HA Offset (deg)'
#            set ylabel 'T antenna (K)'
#            set output '" + directory + filename + ".png'
#            plot '" + directory + filename + ".dat' using 1:2
#        "
#        stdin.write commands
#    }

    Open3.popen3('/usr/bin/octave --silent') { |stdin, stdout, stderr| 
    
        puts "On the graphic, right click the LOWEST and UPPER limits of the EXCLUDED regions for baseline/gaussian fitting."
        logfile.write "On the graphic, right click the LOWEST and UPPER limits of the EXCLUDED regions for baseline/gaussian fitting." + "\n"
        puts " * RIGHT click to finnish excluded regions definition and continue with the rest of calculations"
        logfile.write " * RIGHT click to finnish excluded regions definition and continue with the rest of calculations" + "\n"
        commands = "
            % Load data
            m = load('" + directory + filename + ".dat');
            x = m(:,1);
            y = m(:,2);
            
            % Plot data
            clf
            hold on
            set (gca, 'xminortick', 'on', 'yminortick', 'on');
            plot(x(:,1),y,'@1')
            title('" + header['source'] + " - Band " + band + " (" + date + " " + hms + ")');
            xlabel('HA Offset (deg)');
            ylabel('T antenna (K)');
            grid on
            print('" + directory + filename + ".png','-dpng', '-S800,600')
            
            %%%% Permit to exclude regions for the baseline fitting and removing
            xfiltered = x;
            yfiltered = y;
            disp('Baseline exclusion regions: {')
            do
                [xclick1, yclick1, button1] = ginput(1);
                if (button1 > 1)
                    break
                endif
                disp(strcat('\tLOWEST limit of the EXCLUDED region: x=', num2str(xclick1)))
                hold on
                plot ([xclick1; xclick1], [min(y); max(y)], '-0')
                
                [xclick2, yclick2, button2] = ginput(1);
                % If button2 > 1, it will finnish the region exclusion and the loop will end
                
                disp(strcat('\tUPPER limit of the EXCLUDED region: x=', num2str(xclick2)))
                hold on
                plot ([xclick2; xclick2], [min(y); max(y)], '-0') 
                
                % Draw the area that will be excluded
                area([xclick1,xclick2],[max(y),max(y)],min(y), 'facecolor', 'white')
                
                xclick = [xclick1; xclick2];
                
                % Find positions of the excluded area
                exclude_maxlimit_position = length(find(xfiltered > max(xclick)));
                exclude_minlimit_position = exclude_maxlimit_position + (length(xfiltered) - exclude_maxlimit_position - length(find(xfiltered < min(xclick))));
                xfiltered = [xfiltered(1:exclude_maxlimit_position); xfiltered(exclude_minlimit_position:length(yfiltered))];
                yfiltered = [yfiltered(1:exclude_maxlimit_position); yfiltered(exclude_minlimit_position:length(yfiltered))];
            until (button2 > 1)
            disp('\t}')
            
            % Save an image with the excluded regions
            title('" + header['source'] + " excluded regions - Band " + band + " (" + date + " " + hms + ")');
            print('" + directory + filename + "_baseline_excluded_regions.png','-dpng', '-S800,600')
            
            % Baseline: adjust polynomy
            [pbase, s, mu] = polyfit(xfiltered, yfiltered, 2);
            %pbase = polyfit(xfiltered, yfiltered, 2);
            [ybase, dybase] = polyval(pbase, x, s, mu);
            % Uncertainty associated to the polynomial coefficients
            dpbase = sqrt(sumsq(inv(s.R'))'/s.df)*s.normr;

            % Plot data with baseline
            clf
            set (gca, 'xminortick', 'on', 'yminortick', 'on');
            plot(x(:,1),y,'@1', x,ybase,'-3')
            title('" + header['source'] + " Baseline - Band " + band + " (" + date + " " + hms + ")');
            xlabel('HA Offset (deg)');
            ylabel('T antenna (K)');
            grid on
            print('" + directory + filename + "_baseline.png','-dpng', '-S800,600')
            
            % Calculate residuals (only for selected points)
            [ybasefiltered, dybase] = polyval(pbase, xfiltered, s, mu);
            residuals = yfiltered - ybasefiltered;
            % Plot residuals
            clf
            set (gca, 'xminortick', 'on', 'yminortick', 'on');
            plot(xfiltered(:,1),residuals,'-3')
            title('" + header['source'] + " Baseline residuals - Band " + band + " (" + date + " " + hms + ")');
            xlabel('HA Offset (deg)');
            ylabel('T antenna (K)');
            grid on
            print('" + directory + filename + "_baseline_residuals.png','-dpng', '-S800,600')
            
            % Substract baseline
            y = y - ybase;



            %%%% Permit to exclude regions for the gaussian fitting
            
            % Plot data
            clf
            hold on
            set (gca, 'xminortick', 'on', 'yminortick', 'on');
            plot(x(:,1),y,'@1')
            title('" + header['source'] + " - Band " + band + " (" + date + " " + hms + ")');
            xlabel('HA Offset (deg)');
            ylabel('T antenna (K)');
            grid on
            
            xfiltered = x;
            yfiltered = y;
            disp('Gaussian exclusion regions: {')
            do
                [xclick1, yclick1, button1] = ginput(1);
                if (button1 > 1)
                    break
                endif
                disp(strcat('\tLOWEST limit of the EXCLUDED region: x=', num2str(xclick1)))
                hold on
                plot ([xclick1; xclick1], [min(y); max(y)], '-0')
                
                [xclick2, yclick2, button2] = ginput(1);
                % If button2 > 1, it will finnish the region exclusion and the loop will end
                
                disp(strcat('\tUPPER limit of the EXCLUDED region: x=', num2str(xclick2)))
                hold on
                plot ([xclick2; xclick2], [min(y); max(y)], '-0') 
                
                % Draw the area that will be excluded
                area([xclick1,xclick2],[max(y),max(y)],min(y), 'facecolor', 'white')
                
                xclick = [xclick1; xclick2];
                
                % Find positions of the excluded area
                exclude_maxlimit_position = length(find(xfiltered > max(xclick)));
                exclude_minlimit_position = exclude_maxlimit_position + (length(xfiltered) - exclude_maxlimit_position - length(find(xfiltered < min(xclick))));
                xfiltered = [xfiltered(1:exclude_maxlimit_position); xfiltered(exclude_minlimit_position:length(yfiltered))];
                yfiltered = [yfiltered(1:exclude_maxlimit_position); yfiltered(exclude_minlimit_position:length(yfiltered))];
            until (button2 > 1)
            x = xfiltered;
            y = yfiltered;
            disp('\t}')
            
            % Save an image with the excluded regions
            title('" + header['source'] + " Gaussian excluded regions - Band " + band + " (" + date + " " + hms + ")');
            print('" + directory + filename + "_gaussian_excluded_regions.png','-dpng', '-S800,600')


            % Gaussian function to fit
            function [y] = line_func(x, p)
                % p = vector of parameters to be adjusted
                % x = vector of observations
                y = p(1) * exp(-4 * log(2) * ((x(:,1) - p(2)) / p(3)).**2 );
            end

            % Paramter logical bounds
            n = length(x);
            p1min= 0.0;                 
            p1max= 2.0*max(y);
            
            p2min= 0.8*abs (x(2)-x(1));
            p2max= abs (x(n)-x(1));
            
            p3min= min(x(1),x(n));               
            p3max= max (x(1), x(n));
            
            options.bounds = [p1min, p1max; p2min, p2max; p3min, p3max];
            %options.bounds = [-Inf, Inf; -Inf, Inf; -Inf, Inf];
            
            % Initial parameter's value
            pin=[p1min, p2min, p3max];

            % It is possible to implement the exact derivative of the function to fit or use a generic one (less efficient)
            %dFdp = leasqrdfdp; % exact derivative
            dFdp = 'dfdp';     % estimated derivative

            % Fractional increment of p for numerical partial derivatives
            dp = .001*ones(size(pin));

            % Tolerance on fractional improvement
            stol=.00001;

            % Maximum number of iterations
            niter=20;

            % Statistical weights
            weights = ones(length(y),1);

            [f,p,kvg,iter,corp,covp,covr,stdresid,Z,r2]=leasqr(x,y,pin,'line_func',stol, niter,weights, dp, dFdp, options);
            
            %Uncertainty = Standard deviation
            u = sqrt(diag(covp));

            clf
            xx = [min(x):(max(x)-min(x))/1000:max(x)];
            set (gca, 'xminortick', 'on', 'yminortick', 'on');
            plot(x(:,1),y,'@1', xx,line_func(xx(:),p),'-3')
            title('" + header['source'] + " Gaussian - Band " + band + " (" + date + " " + hms + ")');
            xlabel('HA Offset (deg)');
            ylabel('T antenna (K)');
            grid on
            print('" + directory + filename + "_gaussian.png','-dpng', '-S800,600')
            
            % Plot residuals
            clf
            set (gca, 'xminortick', 'on', 'yminortick', 'on');
            plot(x(:,1),stdresid,'-3')
            title('" + header['source'] +" Gaussian residuals - Band " + band + " (" + date + " " + hms + ")');
            xlabel('HA Offset (deg)');
            ylabel('T antenna (K)');
            grid on
            print('" + directory + filename + "_gaussian_residuals.png','-dpng', '-S800,600')
            
            
            cosdec = cos(" + header["dec"].to_s + " * (pi / 180));
            hpbw = p(3) * cosdec;
            t_ant_hpbw = line_func(hpbw,p);
            t_ant_peak = max(line_func(xx(:),p));
            
            disp('- Baseline polynomy: p(x) = p(1) * x^2 + p(2) * x + p(3)')
            disp('  · Params:')
            disp(strcat('\tp(1) = ', num2str(pbase(1))))
            disp(strcat('\tp(2) = ', num2str(pbase(2))))
            disp(strcat('\tp(3) = ', num2str(pbase(3))))
            disp('  · Uncertainties:')
            disp(strcat('\tq(1) = ', num2str(dpbase(1))))
            disp(strcat('\tq(2) = ', num2str(dpbase(2))))
            disp(strcat('\tq(3) = ', num2str(dpbase(3))))
            disp('- Gaussian: f(x) = p(1) * exp(-4 * log(2) * (x - p(2)) / p(3)).**2')
            disp('  · Params:')
            disp(strcat('\tp(1) = ', num2str(p(1))))
            disp(strcat('\tp(2) = ', num2str(p(2))))
            disp(strcat('\tp(3) = ', num2str(p(3))))
            disp('  · Standard deviation:')
            disp(strcat('\tp(1) = ', num2str(u(1))))
            disp(strcat('\tp(2) = ', num2str(u(2))))
            disp(strcat('\tp(3) = ', num2str(u(3))))
            disp('- Half-power band width:')
            disp(strcat('\tcosdec = ', num2str(cosdec)))
            disp(strcat('\tHPBW = ', num2str(hpbw)))
            disp(strcat('\tAntenna temperature = ', num2str(t_ant_hpbw)))
            disp(strcat('- Antenna temperature peak = ', num2str(t_ant_peak)))
            
            quit
        "
        stdin.write commands
        s = stdout.readlines
        puts s
        logfile.write s
        s = stderr.readlines
        puts s
        logfile.write s
    }
end



